unit jassvmhooks;

{$mode objfpc}{$H+}
{$ASMMODE INTEL}

interface

uses
	hooks, wc3main, wc3jassvm;

type
	TInstanceEventCallback = procedure (hInstance: TInstanceHandle);
	TInstanceExecuteFunctionByNameCallback = function (hInstance: TInstanceHandle; Name: PChar; ScriptResult: PDword): TScriptExitCode;
	TInstanceExecuteFunctionByHandleCallback = function (hInstance: TInstanceHandle; hFunction: Dword; ScriptResult: PDword): TScriptExitCode;
	TInstanceGetStringHandleCallback = function (hInstance: TInstanceHandle; Value: PChar): TJassStringHandle;

var
	pOnInstanceCreated: TInstanceEventCallback;
	pOnInstanceDestroying: TInstanceEventCallback;
	pOnInstanceInit: TInstanceEventCallback;
	pOnExecuteFunctionByName: TInstanceExecuteFunctionByNameCallback;
	pOnExecuteFunctionByHandle: TInstanceExecuteFunctionByHandleCallback;
	pOnInstanceGetStringHandle: TInstanceGetStringHandleCallback;

implementation

var
	orig_RegisterInstance: function ({Instance: PInstance; DesiredSlot: TInstanceHandle}): TInstanceHandle; stdcall; // fastcall
	orig_FreeInstance: procedure ({InstanceId: TInstanceHandle}); stdcall; // fastcall
	orig_Instance_ExecuteFunctionWithOptionalInit: function ({this: PInstance;} FunctionName: PChar; Result: PDword; InitGlobals: LongBool; OpLimit: Dword; Unused: Dword): TScriptExitCode; stdcall; // thiscall
	orig_ExecuteFunctionByHandle: function ({hInstance: TInstanceHandle; hFunction: Dword;} Parameters: Pointer; ScriptResult: PDword; OpLimit: Dword; Unused: Dword; ForceNewRun: LongBool): TScriptExitCode; stdcall; // fastcall
	orig_Instance_GetStringHandle: function ({Instance: PInstance;} Value: PChar): TJassStringHandle; stdcall; // thiscall

function hook_RegisterInstance({Instance: PInstance; DesiredSlot: TInstanceHandle}): TInstanceHandle; stdcall;
begin
	Result:= orig_RegisterInstance();
	pOnInstanceCreated(Result);
end;

procedure hook_FreeInstance({InstanceId: TInstanceHandle}); stdcall; // fastcall
var
	hInstance: TInstanceHandle;
begin
	asm mov hInstance, ecx end;

	pOnInstanceDestroying(hInstance);

	asm mov ecx, hInstance end;
	orig_FreeInstance();
end;

function hook_Instance_ExecuteFunctionWithOptionalInit({this: PInstance;} FunctionName: PChar; ScriptResult: PDword; InitGlobals: LongBool; OpLimit: Dword; Unused: Dword): TScriptExitCode; stdcall; // thiscall
var
	this: PInstance;
begin
	asm mov this, ecx end;

	if InitGlobals then begin
		pOnInstanceInit(this^.Handle);
	end;

	Result:= pOnExecuteFunctionByName(this^.Handle, FunctionName, ScriptResult);
end;

function hook_ExecuteFunctionByHandle({hInstance: TInstanceHandle; hFunction: Dword;} Parameters: Pointer; ScriptResult: PDword; OpLimit: Dword; Unused: Dword; ForceNewRun: LongBool): TScriptExitCode; stdcall; // fastcall
var
	hInstance: TInstanceHandle;
	hFunction: Dword;
begin
	asm
		mov hInstance, ecx
		mov hFunction, edx
	end;

	Result:= pOnExecuteFunctionByHandle(hInstance, hFunction, ScriptResult);
end;

function hook_Instance_GetStringHandle({Instance: PInstance;} Value: PChar): TJassStringHandle; stdcall; // thiscall
var
	this: PInstance;
begin
	asm mov this, ecx end;
	Result:= pOnInstanceGetStringHandle(this^.Handle, Value);
end;

procedure Init;
begin
	Hook(Pointer(_RegisterInstance), @hook_RegisterInstance, Pointer(orig_RegisterInstance));
	Hook(Pointer(_FreeInstance), @hook_FreeInstance, Pointer(orig_FreeInstance));
	Hook(Pointer(_Instance_ExecuteFunctionWithOptionalInit), @hook_Instance_ExecuteFunctionWithOptionalInit, Pointer(orig_Instance_ExecuteFunctionWithOptionalInit));
	Hook(Pointer(_ExecuteFunctionByHandle), @hook_ExecuteFunctionByHandle, Pointer(orig_ExecuteFunctionByHandle));
	Hook(Pointer(_Instance_GetStringHandle), @hook_Instance_GetStringHandle, Pointer(orig_Instance_GetStringHandle));
end;

begin
	wc3main.SubscribeOnInitDone(@Init);
end.

