unit SharedMemory;

{$mode ObjFPC}{$H+}

interface

uses
	Windows;

function CreateSharedConstantString(Name: String; Value: String; out hFileMappingObject: Handle): Boolean;
function GetSharedConstantString(Name: String; out Value: String): Boolean;

implementation

uses
	SysUtils;

type
	TSharedMemoryHeader = record
		Initialized: LongBool;
		Size: Integer;
	end;
	PSharedMemoryHeader = ^TSharedMemoryHeader;

function CreateSharedConstantString(Name: String; Value: String; out hFileMappingObject: Handle): Boolean;
var
	pHeader: PSharedMemoryHeader = nil;
begin
	Result:= False;
	hFileMappingObject:= 0;

	try
		hFileMappingObject:= CreateFileMappingW(
			INVALID_HANDLE_VALUE,
			nil,
			PAGE_READWRITE,
			0,
			SizeOf(TSharedMemoryHeader) + Length(Value),
			PWideChar(UnicodeString(Name))
		);
		if hFileMappingObject = 0 then exit;

		pHeader:= MapViewOfFile(
			hFileMappingObject,
			FILE_MAP_WRITE,
			0,
			0,
			0
		);
		if pHeader = nil then exit;

		pHeader^.Size:= Length(Value);

		if Length(Value) > 0 then begin
			Move(
				Value[1],
				PChar(pHeader + SizeOf(TSharedMemoryHeader))^,
				Length(Value)
			);
		end;

		WriteBarrier;

		pHeader^.Initialized:= True;
		Result:= True;
	finally
		if pHeader <> nil
			then UnmapViewOfFile(pHeader);

		if (Result = False) and (hFileMappingObject <> 0)
			then CloseHandle(hFileMappingObject);
	end;
end;

function GetSharedConstantString(Name: String; out Value: String): Boolean;
var
	hFileMappingObject: Handle = 0;
	pHeader: PSharedMemoryHeader = nil;
begin
	Result:= False;
	Value:= '';

	try
		hFileMappingObject:= OpenFileMappingW(
			FILE_MAP_READ,
			False,
			PWideChar(UnicodeString(Name))
		);
		if hFileMappingObject = 0 then exit;

		pHeader:= MapViewOfFile(hFileMappingObject, FILE_MAP_READ, 0, 0, 0);
		if pHeader = nil then exit;

		if not pHeader^.Initialized then exit;

		if pHeader^.Size > 0 then begin
			SetLength(Value, pHeader^.Size);
			Move(
				PChar(pHeader + SizeOf(TSharedMemoryHeader))^,
				Value[1],
				Length(Value)
			);
		end;

		Result:= True;
	finally
		if pHeader <> nil
			then UnmapViewOfFile(pHeader);

		if hFileMappingObject <> 0
			then CloseHandle(hFileMappingObject);
	end;

end;

end.

